@@ -32,7 +32,10 @@ from gnomemusic.query import Query
 from gnomemusic import log, TrackerWrapper
 import logging
 import os
-os.environ['GRL_PLUGIN_RANKS'] = 'grl-local-metadata:4,grl-filesystem:3,grl-tracker-source:2,grl-lastfm-cover:1'
+os.environ['GRL_PLUGIN_RANKS'] = ("grl-local-metadata:5,"
+                                  "grl-filesystem:4,"
+                                  "grl-tracker-source:3,"
+                                  "grl-lastfm-cover:2")
 from gi.repository import Grl
 logger = logging.getLogger(__name__)
 
@@ -46,13 +49,22 @@ class Grilo(GObject.GObject):
     }
 
     METADATA_KEYS = [
-        Grl.METADATA_KEY_ID, Grl.METADATA_KEY_TITLE,
-        Grl.METADATA_KEY_ARTIST, Grl.METADATA_KEY_ALBUM,
-        Grl.METADATA_KEY_DURATION,
+        Grl.METADATA_KEY_ALBUM,
+        Grl.METADATA_KEY_ALBUM_ARTIST,
+        Grl.METADATA_KEY_ALBUM_DISC_NUMBER,
+        Grl.METADATA_KEY_ARTIST,
         Grl.METADATA_KEY_CREATION_DATE,
-        Grl.METADATA_KEY_URL,
+        Grl.METADATA_KEY_COMPOSER,
+        Grl.METADATA_KEY_DURATION,
+        Grl.METADATA_KEY_FAVOURITE,
+        Grl.METADATA_KEY_ID,
         Grl.METADATA_KEY_LYRICS,
-        Grl.METADATA_KEY_THUMBNAIL]
+        Grl.METADATA_KEY_PLAY_COUNT,
+        Grl.METADATA_KEY_THUMBNAIL,
+        Grl.METADATA_KEY_TITLE,
+        Grl.METADATA_KEY_TRACK_NUMBER,
+        Grl.METADATA_KEY_URL
+    ]
 
     METADATA_THUMBNAIL_KEYS = [
         Grl.METADATA_KEY_ID,
@@ -83,7 +95,13 @@ class Grilo(GObject.GObject):
                                                Grl.ResolutionFlags.IDLE_RELAY)
 
         self.sources = {}
-        self.blacklist = ['grl-filesystem', 'grl-bookmarks', 'grl-metadata-store', 'grl-podcasts']
+        self.blacklist = [
+            'grl-filesystem',
+            'grl-bookmarks',
+            'grl-metadata-store',
+            'grl-podcasts',
+            'grl-spotify-cover'
+        ]
         self.tracker = None
         self.changed_media_ids = []
         self.pending_event_id = 0
@@ -120,10 +138,11 @@ class Grilo(GObject.GObject):
                     media_id = media.get_id()
                     if changeType == Grl.SourceChangeType.ADDED:
                         # Check that this media is an audio file
-                        query = "select DISTINCT rdf:type nie:mimeType(?urn) as mime-type" +\
-                                " { ?urn rdf:type nie:InformationElement . FILTER (tracker:id(?urn) = %s) }" % media_id
-                        mimeType = grilo.tracker.query_sync(query, [Grl.METADATA_KEY_MIME], grilo.options)[0].get_mime()
-                        if mimeType and mimeType.startswith("audio"):
+                        mime_type = self.tracker.query_sync(
+                            Query.is_audio(media_id),
+                            [Grl.METADATA_KEY_MIME],
+                            self.options)[0].get_mime()
+                        if mime_type and mime_type.startswith("audio"):
                             self.changed_media_ids.append(media_id)
                     if changeType == Grl.SourceChangeType.REMOVED:
                         # There is no way to check that removed item is a media
@@ -172,6 +191,15 @@ class Grilo(GObject.GObject):
 
     @log
     def _on_source_added(self, pluginRegistry, mediaSource):
+        if ("net:plaintext" in mediaSource.get_tags()
+                or mediaSource.get_id() in self.blacklist):
+            try:
+                pluginRegistry.unregister_source(mediaSource)
+            except GLib.GError:
+                logger.error("Failed to unregister %s.",
+                             mediaSource.get_id())
+            return
+
         id = mediaSource.get_id()
         logger.debug("new grilo source %s was added", id)
         try:
@@ -196,8 +224,8 @@ class Grilo(GObject.GObject):
                 self.sources[id] = mediaSource
                 self.emit('new-source-added', mediaSource)
 
-            elif (id not in self.blacklist) and (ops & Grl.SupportedOps.SEARCH) and \
-                 (mediaSource.get_supported_media() & Grl.MediaType.AUDIO):
+            elif (ops & Grl.SupportedOps.SEARCH
+                  and mediaSource.get_supported_media() & Grl.MediaType.AUDIO):
                 logger.debug("source %s is searchable", id)
                 self.sources[id] = mediaSource
                 self.emit('new-source-added', mediaSource)
@@ -211,34 +239,46 @@ class Grilo(GObject.GObject):
 
     @log
     def populate_artists(self, offset, callback, count=-1):
-        self.populate_items(Query.all_artists(), offset, callback, count)
+        if self.tracker:
+            GLib.idle_add(self.populate_items, Query.all_artists(), offset,
+                          callback, count)
 
     @log
     def populate_albums(self, offset, callback, count=-1):
-        self.populate_items(Query.all_albums(), offset, callback, count)
+        if self.tracker:
+            GLib.idle_add(self.populate_items, Query.all_albums(), offset,
+                                                callback, count)
 
     @log
     def populate_songs(self, offset, callback, count=-1):
-        self.populate_items(Query.all_songs(), offset, callback, count)
+        if self.tracker:
+            GLib.idle_add(self.populate_items, Query.all_songs(), offset,
+                                                callback, count)
 
     @log
     def populate_playlists(self, offset, callback, count=-1):
-        self.populate_items(Query.all_playlists(), offset, callback, count)
+        if self.tracker:
+            GLib.idle_add(self.populate_items, Query.all_playlists(), offset,
+                                                callback, count)
 
     @log
     def populate_album_songs(self, album, callback, count=-1):
         if album.get_source() == 'grl-tracker-source':
-            self.populate_items(Query.album_songs(album.get_id()), 0, callback, count)
+            GLib.idle_add(self.populate_items,
+                          Query.album_songs(album.get_id()), 0, callback, count)
         else:
             source = self.sources[album.get_source()]
-            length = len(album.tracks)
-            for i, track in enumerate(album.tracks):
+            length = len(album.songs)
+            for i, track in enumerate(album.songs):
                 callback(source, None, track, length - (i + 1), None)
             callback(source, None, None, 0, None)
 
     @log
     def populate_playlist_songs(self, playlist, callback, count=-1):
-        self.populate_items(Query.playlist_songs(playlist.get_id()), 0, callback, count)
+        if self.tracker:
+            GLib.idle_add(self.populate_items,
+                          Query.playlist_songs(playlist.get_id()), 0, callback,
+                          count)
 
     @log
     def populate_custom_query(self, query, callback, count=-1, data=None):
@@ -257,14 +297,34 @@ class Grilo(GObject.GObject):
 
     @log
     def toggle_favorite(self, song_item):
-        # TODO: change "bool(song_item.get_lyrics())" --> song_item.get_favourite() once query works properly
-        # TODO: when .set/get_favourite work, set_favourite outside loop: item.set_favourite(!item.get_favourite())
-        if bool(song_item.get_lyrics()):  # is favorite
-            self.sparqltracker.update(Query.remove_favorite(song_item.get_url()), GLib.PRIORITY_DEFAULT, None)
+        """Toggles favorite status for media item
+
+        Toggles favorite status and writes it back to the tracker store
+        :param song_item: A Grilo media item
+        """
+        if song_item.get_favourite():
+            # For now keep unsetting the lyrics to deal with how
+            # previous versions dealt with favorites.
             song_item.set_lyrics("")
-        else:  # not favorite
-            self.sparqltracker.update(Query.add_favorite(song_item.get_url()), GLib.PRIORITY_DEFAULT, None)
-            song_item.set_lyrics("i'm truthy")
+            song_item.set_favourite(False)
+        else:
+            song_item.set_favourite(True)
+
+        # FIXME: We assume this is the tracker plugin.
+        # FIXME: Doing this async crashes
+        self.tracker.store_metadata_sync(song_item,
+                                         [Grl.METADATA_KEY_FAVOURITE],
+                                         Grl.WriteFlags.NORMAL)
+
+    @log
+    def set_favorite(self, song_item, favorite):
+        """Set the favorite status of a media item
+
+        :param song_item: A Grilo media item
+        :param bool favorite: Set favorite status
+        """
+        if song_item.get_favourite() != favorite:
+            self.toggle_favorite(song_item)
 
     @log
     def search(self, q, callback, data=None):
@@ -292,10 +352,9 @@ class Grilo(GObject.GObject):
                                 _multiple_search_callback, data)
 
     @log
-    def get_album_art_for_item(self, item, callback, data=None):
+    def get_album_art_for_item(self, item, callback):
         item_id = item.get_id()
 
-        query = None
         if item.is_audio():
             query = Query.get_album_for_song_id(item_id)
         else:
@@ -304,7 +363,8 @@ class Grilo(GObject.GObject):
         options = self.full_options.copy()
         options.set_count(1)
 
-        self.search_source.query(query, self.METADATA_THUMBNAIL_KEYS, options, callback, data)
+        self.search_source.query(query, self.METADATA_THUMBNAIL_KEYS, options,
+                                 callback)
 
     @log
     def get_playlist_with_id(self, playlist_id, callback):
@@ -320,5 +380,113 @@ class Grilo(GObject.GObject):
 
         self.tracker.query(query, self.METADATA_KEYS, options, callback, None)
 
+    @log
+    def bump_play_count(self, media):
+        """Bumps the play count of a song
+
+        Adds one to the playcount and adds it to the tracker store
+        :param media: A Grilo media item
+        """
+        count = media.get_play_count()
+        media.set_play_count(count + 1)
+
+        # FIXME: We assume this is the tracker plugin.
+        # FIXME: Doing this async crashes
+        self.tracker.store_metadata_sync(media, [Grl.METADATA_KEY_PLAY_COUNT],
+                                         Grl.WriteFlags.NORMAL)
+
+    @log
+    def set_last_played(self, media):
+        """Sets the date-time when the media was last played
+
+        Sets the last played date-time for the media.
+        :param media: A Grilo media item
+        """
+        media.set_last_played(GLib.DateTime.new_now_utc())
+        # FIXME: We assume this is the tracker plugin.
+        # FIXME: Doing this async crashes
+        self.tracker.store_metadata_sync(media, [Grl.METADATA_KEY_LAST_PLAYED],
+                                         Grl.WriteFlags.NORMAL)
+
+    @log
+    def songs_available(self, callback):
+        """Checks if there are any songs available
+
+        Calls a callback function with True or False depending on the
+        availability of songs.
+        :param callback: Function to call on result
+        """
+        def cursor_next_cb(conn, res, data):
+            try:
+                has_next = conn.next_finish(res)
+            except GLib.Error as err:
+                logger.warn("Error: %s, %s", err.__class__, err)
+                callback(False)
+                return
+
+            if has_next:
+                count = conn.get_integer(0)
+
+                if count > 0:
+                    callback(True)
+                    return
+
+            callback(False)
+
+        def songs_query_cb(conn, res, data):
+            try:
+                cursor = conn.query_finish(res)
+            except GLib.Error as err:
+                logger.warn("Error: %s, %s", err.__class__, err)
+                callback(False)
+                return
+
+            cursor.next_async(None, cursor_next_cb, None)
+
+        # TODO: currently just checks tracker, should work with any
+        # queryable supported Grilo source.
+        self.sparqltracker.query_async(Query.all_songs_count(), None,
+                                       songs_query_cb, None)
+
+    @log
+    def playlists_available(self, callback):
+        """Checks if there are any non-static playlists available
+
+        Calls a callback function with True or False depending on the
+        availability of playlists.
+        :param callback: Function to call on result
+        """
+        def cursor_next_cb(conn, res, data):
+            try:
+                has_next = conn.next_finish(res)
+            except GLib.Error as err:
+                logger.warn("Error: %s, %s", err.__class__, err)
+                callback(False)
+                return
+
+            if has_next:
+                count = conn.get_integer(0)
+
+                if count > 0:
+                    callback(True)
+                    return
+
+            callback(False)
+
+        def playlists_query_cb(conn, res, data):
+            try:
+                cursor = conn.query_finish(res)
+            except GLib.Error as err:
+                logger.warn("Error: %s, %s", err.__class__, err)
+                callback(False)
+                return
+
+            cursor.next_async(None, cursor_next_cb, None)
+
+        # TODO: currently just checks tracker, should work with any
+        # queryable supported Grilo source.
+        self.sparqltracker.query_async(
+                Query.all_non_static_playlists_count(),
+                None, playlists_query_cb, None)
 
 grilo = Grilo()
